"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var path = require("path");
var crypto = require("crypto");
function hash(str) {
    var h = crypto.createHash('sha256');
    h.update(str);
    return h.digest('hex');
}
function getId(symbol) {
    return symbol.id;
}
function apiTransformer(program) {
    function visitNodeAndChildren(node, program, context) {
        return ts.visitEachChild(visitNode(node, program), function (childNode) { return visitNodeAndChildren(childNode, program, context); }, context);
    }
    function visitNode(node, program) {
        var typeChecker = program.getTypeChecker();
        if (isCallExpression(node, typeChecker)) {
            var methodName = getCallExpressionName(node, typeChecker);
            if (methodName === 'InterfaceSymbol') {
                return visitSymbolNode(node, typeChecker);
            }
            else if (methodName === 'bind' || methodName === 'resolve') {
                return visitParametersNode(node, typeChecker);
            }
        }
        return node;
    }
    function visitSymbolNode(node, typeChecker) {
        if (!node.typeArguments) {
            throw new Error("No type arguments provided to `InterfaceSymbol<T>()`.");
        }
        var typeArgument = node.typeArguments[0];
        var type = typeChecker.getTypeFromTypeNode(typeArgument);
        var symbol = type.symbol;
        if (type.isClass() || !type.isClassOrInterface()) {
            throw new Error("The type provided is not an interface.");
        }
        var uid = symbol.escapedName + "#" + hash(prefix + getId(symbol));
        return ts.createCall(ts.createIdentifier('Symbol.for'), [], [ts.createStringLiteral(uid)]);
    }
    function getImports(root) {
        var fake = root;
        if (fake.imports && Array.isArray(fake.imports)) {
            return fake.imports;
        }
        else {
            return [];
        }
    }
    function getImportForType(typeChecker, imports, type) {
        for (var _i = 0, imports_1 = imports; _i < imports_1.length; _i++) {
            var node = imports_1[_i];
            if (ts.isImportDeclaration(node.parent)) {
                var importClause = node.parent.importClause;
                if (importClause && importClause.namedBindings && ts.isNamedImports(importClause.namedBindings)) {
                    var elements = importClause.namedBindings.elements;
                    for (var _a = 0, elements_1 = elements; _a < elements_1.length; _a++) {
                        var element = elements_1[_a];
                        if (ts.isImportSpecifier(element)) {
                            var importType = typeChecker.getTypeFromTypeNode(element.name);
                            var paramType = typeChecker.getTypeFromTypeNode(type);
                            if (importType.id === paramType.id) {
                                return element;
                            }
                        }
                    }
                }
            }
        }
        return undefined;
    }
    function visitParametersNode(node, typeChecker) {
        var signature = typeChecker.getResolvedSignature(node);
        if (!signature) {
            return node;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return node;
        }
        if (!ts.isMethodDeclaration(declaration) || !declaration.name) {
            return node;
        }
        var numArgs = declaration.name.getText() === 'bind' ? 2 : 1;
        if (node.arguments.length !== numArgs) {
            return node;
        }
        var type = typeChecker.getTypeFromTypeNode(node.arguments[numArgs - 1]);
        if (!type.symbol.members)
            return node;
        var params = [];
        type.symbol.members.forEach(function (val) {
            if (val.declarations.length === 0)
                return;
            var firstDeclaration = val.declarations[0];
            if (ts.isConstructorDeclaration(firstDeclaration)) {
                for (var _i = 0, _a = firstDeclaration.parameters; _i < _a.length; _i++) {
                    var param = _a[_i];
                    if (param.type) {
                        var isArray = false;
                        var type_1 = typeChecker.getTypeFromTypeNode(param.type);
                        if (ts.isArrayTypeNode(param.type)) {
                            var t = type_1;
                            if (t.typeArguments && t.typeArguments.length > 0) {
                                type_1 = t.typeArguments[0];
                                isArray = true;
                            }
                        }
                        var symbol = type_1.symbol;
                        if ((type_1.isClassOrInterface() && !type_1.isClass())) {
                            var uid = symbol.escapedName + "#" + hash(prefix + getId(symbol));
                            params.push(ts.createCall(ts.createIdentifier('Symbol.for'), [], [ts.createStringLiteral(uid)]));
                        }
                        else {
                            var imports = getImports(node.getSourceFile());
                            var importElement = getImportForType(typeChecker, imports, param.type);
                            if (importElement) {
                                params.push(importElement.name);
                            }
                            else {
                                params.push(ts.createLiteral(symbol.name));
                            }
                        }
                        if (isArray) {
                            var v = params.pop();
                            if (v) {
                                params.push(ts.createArrayLiteral([v]));
                            }
                        }
                    }
                }
            }
        });
        var args = [];
        for (var _i = 0, _a = node.arguments; _i < _a.length; _i++) {
            var val = _a[_i];
            args.push(val);
        }
        args.push(ts.createArrayLiteral(params));
        node.arguments = ts.createNodeArray(args);
        return node;
    }
    var apiTs = path.dirname(__dirname).replace(/\\/g, "/");
    function getCallExpressionName(node, typeChecker) {
        if (!ts.isCallExpression(node)) {
            return undefined;
        }
        var signature = typeChecker.getResolvedSignature(node);
        if (typeof signature === 'undefined') {
            return undefined;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return undefined;
        }
        var dirname = path.dirname(declaration.getSourceFile().fileName).replace(/\\/g, "/");
        if (apiTs !== dirname && dirname.indexOf(apiTs) !== 0) {
            return undefined;
        }
        if ((ts.isMethodDeclaration(declaration) || ts.isFunctionDeclaration(declaration)) && !!declaration.name) {
            return declaration.name.getText();
        }
        return undefined;
    }
    function isCallExpression(node, typeChecker) {
        if (node.kind !== ts.SyntaxKind.CallExpression) {
            return false;
        }
        var signature = typeChecker.getResolvedSignature(node);
        if (typeof signature === 'undefined') {
            return false;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return false;
        }
        return (ts.isMethodDeclaration(declaration)
            || ts.isFunctionDeclaration(declaration)) && !!declaration.name;
    }
    var prefix = crypto.randomBytes(256).toString('hex');
    return function (context) { return function (file) { return visitNodeAndChildren(file, program, context); }; };
}
exports.default = apiTransformer;
//# sourceMappingURL=transformer.js.map